/*
 * Copyright © Original Authors
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Additional copyright and licensing notices may apply for content that was
 * included from other projects. For more information, see ATTRIBUTION.md.
 */

package grondag.xm.api.primitive.simple;

import org.jetbrains.annotations.ApiStatus.Experimental;
import io.vram.frex.api.model.util.ColorUtil;
import net.minecraft.class_2350;
import net.minecraft.class_2350.class_2351;
import net.minecraft.class_2350.class_2352;
import net.minecraft.class_2382;
import grondag.fermion.orientation.api.OrientationType;
import grondag.xm.Xm;
import grondag.xm.api.connect.state.SimpleJoinState;
import grondag.xm.api.mesh.CsgMeshBuilder;
import grondag.xm.api.mesh.WritableMesh;
import grondag.xm.api.mesh.XmMesh;
import grondag.xm.api.mesh.polygon.FaceVertex;
import grondag.xm.api.mesh.polygon.MutablePolygon;
import grondag.xm.api.mesh.polygon.PolyHelper;
import grondag.xm.api.mesh.polygon.PolyTransform;
import grondag.xm.api.mesh.polygon.Polygon;
import grondag.xm.api.modelstate.primitive.PrimitiveState;
import grondag.xm.api.paint.SurfaceTopology;
import grondag.xm.api.primitive.SimplePrimitive;
import grondag.xm.api.primitive.surface.XmSurface;
import grondag.xm.api.primitive.surface.XmSurfaceList;
import grondag.xm.api.texture.TextureOrientation;

@Experimental
public class CappedSquareInsetColumn {
	private CappedSquareInsetColumn() { }

	public static final XmSurfaceList SURFACES = XmSurfaceList.builder()
		.add("ends", SurfaceTopology.CUBIC, XmSurface.FLAG_NONE)
		.add("outer", SurfaceTopology.CUBIC, XmSurface.FLAG_NONE)
		.add("cut", SurfaceTopology.CUBIC, XmSurface.FLAG_LAMP_GRADIENT)
		.add("inner", SurfaceTopology.CUBIC, XmSurface.FLAG_LAMP)
		.build();

	public static final XmSurface SURFACE_ENDS = SURFACES.get(0);
	public static final XmSurface SURFACE_OUTER = SURFACES.get(1);
	public static final XmSurface SURFACE_CUT = SURFACES.get(2);
	public static final XmSurface SURFACE_INNER = SURFACES.get(3);

	private static class_2351[] AXES = class_2351.values();

	static final float MIN = 2f/16f;
	static final float MAX = 1 - MIN;
	static final float DEPTH = 1f/8f;
	static final float IMIN = MIN + DEPTH;
	static final float IMAX = 1 - IMIN;

	private static XmMesh factory(PrimitiveState modelState) {
		final PolyTransform pt = PolyTransform.get(modelState);

		final CsgMeshBuilder csg = CsgMeshBuilder.threadLocal();
		final SimpleJoinState state = modelState.simpleJoin();

		boolean hasUpper = false;
		boolean hasLower = false;
		final class_2351 axis = AXES[modelState.orientationIndex()];
		final class_2350 up = class_2350.method_10169(axis, class_2352.field_11056);
		final class_2350 down = class_2350.method_10169(axis, class_2352.field_11060);

		emitMiddleSection(csg, pt, modelState.primitive().lampSurface(modelState) != null);

		csg.push();

		if (!state.isJoined(up)) {
			emitCapSection(csg.input(), pt, axis != class_2351.field_11048);
			csg.union();
			hasUpper = true;
		}

		if (!state.isJoined(down)) {
			emitCapSection(csg.input(), pt, axis == class_2351.field_11048);
			csg.union();
			hasLower = true;
		}

		csg.pop();

		if (hasLower || hasUpper) {
			csg.union();
		}

		// Filter out quads that are occluded by connection
		final WritableMesh result = csg.buildMutable();

		if (!hasLower || !hasUpper) {
			final Polygon reader = result.reader();

			if (reader.origin()) {
				do {
					final class_2350 cullFace = reader.cullFace();

					if ((!hasLower && cullFace == down) || (!hasUpper && cullFace == up)) {
						reader.delete();
					}
				} while (reader.next());
			}
		}

		return result.releaseToReader();
	}

	private static void emitMiddleSection(CsgMeshBuilder csg, PolyTransform pt, boolean isLit) {
		final MutablePolygon writer = csg.input().writer();

		writer
			.colorAll(0, 0xFFFFFFFF)
			.surface(SURFACE_OUTER)
			.lockUV(0, false)
			.rotation(0, TextureOrientation.IDENTITY)
			.sprite(0, "")
			.saveDefaults();

		writer.setupFaceQuad(class_2350.field_11033, MIN, MIN, MAX, MAX, 0, class_2350.field_11043);
		pt.accept(writer);
		writer.append();

		writer.setupFaceQuad(class_2350.field_11036, MIN, MIN, MAX, MAX, 0, class_2350.field_11043);
		pt.accept(writer);
		writer.append();

		writer.setupFaceQuad(class_2350.field_11034, MIN, 0, MAX, 1, MIN, class_2350.field_11036);
		pt.accept(writer);
		writer.append();

		writer.setupFaceQuad(class_2350.field_11039, MIN, 0, MAX, 1, MIN, class_2350.field_11036);
		pt.accept(writer);
		writer.append();

		writer.setupFaceQuad(class_2350.field_11043, MIN, 0, MAX, 1, MIN, class_2350.field_11036);
		pt.accept(writer);
		writer.append();

		writer.setupFaceQuad(class_2350.field_11035, MIN, 0, MAX, 1, MIN, class_2350.field_11036);
		pt.accept(writer);
		writer.append();

		csg.union();

		cutSide(class_2350.field_11034, csg, pt, isLit);
		cutSide(class_2350.field_11039, csg, pt, isLit);
		cutSide(class_2350.field_11043, csg, pt, isLit);
		cutSide(class_2350.field_11035, csg, pt, isLit);
	}

	// UGLY: some copy pasta
	private static void cutSide(class_2350 face, CsgMeshBuilder csg, PolyTransform pt, boolean isLit) {
		final WritableMesh mesh = csg.input();
		final MutablePolygon writer = mesh.writer();

		writer.colorAll(0, 0xFFFFFFFF);
		writer.lockUV(0, true);
		writer.rotation(0, TextureOrientation.IDENTITY);
		writer.sprite(0, "");
		writer.saveDefaults();

		writer.surface(SURFACE_INNER);
		writer.setupFaceQuad(face.method_10153(), IMIN, 0, IMAX, 1, 1 - IMIN, class_2350.field_11036);
		pt.accept(writer);
		writer.append();

		writer.surface(SURFACE_OUTER);
		writer.setupFaceQuad(face, IMIN, 0, IMAX, 1, 0, class_2350.field_11036);
		pt.accept(writer);
		writer.append();

		setupCutSideQuad(writer, pt, IMIN, 1 - IMIN, 1 - IMIN, 1, 0, class_2350.field_11033, face, isLit);
		setupCutSideQuad(writer, pt, IMIN, 1 - IMIN, 1 - IMIN, 1, 0, class_2350.field_11036, face, isLit);

		setupCutSideQuad(writer, pt, 0, 1 - IMIN, 1, 1, IMIN, PolyHelper.leftOf(face, class_2350.field_11036), face, isLit);
		setupCutSideQuad(writer, pt, 0, 1 - IMIN, 1, 1, IMIN, PolyHelper.rightOf(face, class_2350.field_11036), face, isLit);
		csg.difference();
	}

	// UGLY: copy pasta
	private static void setupCutSideQuad(MutablePolygon poly, PolyTransform pt, float x0, float y0, float x1, float y1, float depth, class_2350 face, class_2350 topFace, boolean isLit) {
		final int glow = isLit ? 255 : 0;

		poly.surface(SURFACE_CUT);

		poly.setupFaceQuad(
			face,
			new FaceVertex.Colored(x0, y0, depth, ColorUtil.WHITE, glow),
			new FaceVertex.Colored(x1, y0, depth, ColorUtil.WHITE, glow),
			new FaceVertex.Colored(x1, y1, depth, ColorUtil.WHITE, glow / 3),
			new FaceVertex.Colored(x0, y1, depth, ColorUtil.WHITE, glow / 3),
			topFace
		);

		// force vertex normals out to prevent lighting anomalies
		final class_2382 vec = face.method_10163();
		final float x = vec.method_10263();
		final float y = vec.method_10264();
		final float z = vec.method_10260();

		for (int i = 0; i < 4; i++) {
			poly.normal(i, x, y, z);
		}

		pt.accept(poly);
		poly.append();
	}

	private static void emitCapSection(WritableMesh mesh, PolyTransform pt, boolean top) {
		final MutablePolygon writer = mesh.writer();

		writer
			.colorAll(0, 0xFFFFFFFF)
			.surface(SURFACE_ENDS)
			.lockUV(0, true)
			.rotation(0, TextureOrientation.IDENTITY)
			.sprite(0, "")
			.saveDefaults();

		writer.setupFaceQuad(class_2350.field_11033, 0, 0, 1, 1, top ? 0.75f : 0, class_2350.field_11043);
		pt.accept(writer);
		writer.append();

		writer.setupFaceQuad(class_2350.field_11036, 0, 0, 1, 1, top ? 0 : 0.75f, class_2350.field_11043);
		pt.accept(writer);
		writer.append();

		final float min = top ? .75f : 0;
		final float max = min + 0.25f;

		writer.setupFaceQuad(class_2350.field_11034, 0, min, 1, max, 0, class_2350.field_11036);
		pt.accept(writer);
		writer.append();

		writer.setupFaceQuad(class_2350.field_11039, 0, min, 1, max, 0, class_2350.field_11036);
		pt.accept(writer);
		writer.append();

		writer.setupFaceQuad(class_2350.field_11043, 0, min, 1, max, 0, class_2350.field_11036);
		pt.accept(writer);
		writer.append();

		writer.setupFaceQuad(class_2350.field_11035, 0, min, 1, max, 0, class_2350.field_11036);
		pt.accept(writer);
		writer.append();
	}

	public static final SimplePrimitive INSTANCE = SimplePrimitive.builder()
		.surfaceList(SURFACES)
		.polyFactory(CappedSquareInsetColumn::factory)
		.axisJoin(true)
		.orientationType(OrientationType.AXIS)
		.build(Xm.id("capped_square_inset_column"));
}
