/*
 * Copyright © Original Authors
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Additional copyright and licensing notices may apply for content that was
 * included from other projects. For more information, see ATTRIBUTION.md.
 */

package grondag.xm.painter;

import it.unimi.dsi.fastutil.HashCommon;
import net.minecraft.class_2350;
import net.minecraft.class_2382;
import org.jetbrains.annotations.ApiStatus.Internal;
import grondag.fermion.orientation.api.ClockwiseRotation;
import grondag.fermion.varia.Useful;
import grondag.xm.api.mesh.MutableMesh;
import grondag.xm.api.mesh.polygon.MutablePolygon;
import grondag.xm.api.modelstate.base.BaseModelState;
import grondag.xm.api.paint.XmPaint;
import grondag.xm.api.primitive.surface.XmSurface;
import grondag.xm.api.texture.TextureOrientation;
import grondag.xm.api.texture.TextureScale;
import grondag.xm.api.texture.TextureSet;

@Internal
@SuppressWarnings("rawtypes")
public abstract class AbstractQuadPainter {
	@FunctionalInterface
	public interface PaintMethod {
		/**
		 * Assigns specific texture and texture rotation based on model state and
		 * information in the polygon and surface. Also handles texture UV mapping.
		 *
		 * <p>Implementations can and should assume locked UV coordinates are assigned
		 * before this is called if UV locking is enabled for the quad
		 *
		 * <p>Implementation should claim and use first render layer with a null texture
		 * name. (Claim by assigning a non-null texture name.)
		 *
		 * <p>Any polys in the input stream that are split should be deleted and new polys
		 * appended to the stream.
		 *
		 * <p>Implementation may assume stream is non-empty and stream editor is at origin.
		 */
		void paintQuads(MutableMesh stream, BaseModelState modelState, XmSurface surface, XmPaint paint, int textureDepth);
	}

	/**
	 * Call from paint quad in sub classes to return results. Handles item scaling,
	 * then adds to the output list.
	 */
	protected static void commonPostPaint(MutablePolygon editor, BaseModelState modelState, XmSurface surface, XmPaint paint, int textureIndex) {
		if (textureIndex == 0) {
			editor.blendMode(paint.blendMode());
		}

		editor.emissive(textureIndex, paint.emissive(textureIndex));
		editor.disableAo(textureIndex, paint.disableAo(textureIndex));
		editor.disableDiffuse(textureIndex, paint.disableDiffuse(textureIndex));
		paint.vertexProcessor(textureIndex).process(editor, modelState, surface, paint, textureIndex);
	}

	/**
	 * Transform input vector so that x & y correspond with u / v on the given face,
	 * with u,v origin at upper left and z is depth, where positive values represent
	 * distance into the face (away from viewer).
	 *
	 * <p>Coordinates are start masked to the scale of the texture being used and when
	 * we reverse an orthogonalAxis, we use the texture's sliceMask as the basis so
	 * that we remain within the frame of the texture scale we are using.
	 *
	 * <p>Note that the x, y components are for determining min/max UV values. They
	 * should NOT be used to set vertex UV coordinates directly. All bigtex models
	 * should have lockUV = true, which means that uv coordinates will be derived at
	 * time of quad bake by projecting each vertex onto the plane of the quad's
	 * nominal face. Setting UV coordinates on a quad with lockUV=true has no
	 * effect.
	 */
	protected static class_2382 getSurfaceVector(int blockX, int blockY, int blockZ, class_2350 face, TextureScale scale) {
		// PERF: reuse instances?

		final int sliceCountMask = scale.sliceCountMask;
		final int x = blockX & sliceCountMask;
		final int y = blockY & sliceCountMask;
		final int z = blockZ & sliceCountMask;

		switch (face) {
			case field_11034:
				return new class_2382(sliceCountMask - z, sliceCountMask - y, -blockX);

			case field_11039:
				return new class_2382(z, sliceCountMask - y, blockX);

			case field_11043:
				return new class_2382(sliceCountMask - x, sliceCountMask - y, blockZ);

			case field_11035:
				return new class_2382(x, sliceCountMask - y, -blockZ);

			case field_11033:
				return new class_2382(x, sliceCountMask - z, blockY);

			case field_11036:
			default:
				return new class_2382(x, z, -blockY);
		}
	}

	/**
	 * Rotates given surface vector around the center of the texture by the given
	 * degree.
	 *
	 */
	protected static class_2382 rotateFacePerspective(class_2382 vec, ClockwiseRotation rotation, TextureScale scale) {
		// PERF - reuse instances?
		switch (rotation) {
			case ROTATE_90:
				return new class_2382(vec.method_10264(), scale.sliceCountMask - vec.method_10263(), vec.method_10260());

			case ROTATE_180:
				return new class_2382(scale.sliceCountMask - vec.method_10263(), scale.sliceCountMask - vec.method_10264(), vec.method_10260());

			case ROTATE_270:
				return new class_2382(scale.sliceCountMask - vec.method_10264(), vec.method_10263(), vec.method_10260());

			case ROTATE_NONE:
			default:
				return vec;
		}
	}

	protected static int textureVersionForFace(class_2350 face, TextureSet tex, BaseModelState modelState) {
		return tex.versionCount() == 0 ? 0 : textureHashForFace(face, tex, modelState) & tex.versionMask();
	}

	protected static int textureHashForFace(class_2350 face, TextureSet tex, BaseModelState modelState) {
		final int species = modelState.hasSpecies() ? modelState.species() : 0;
		final int speciesBits = species << 16;
		final int shift = tex.scale().power;

		switch (face) {
			case field_11033:
			case field_11036: {
				final int yBits = (((modelState.posX() >> shift) & 0xFF) << 8) | ((modelState.posZ() >> shift) & 0xFF) | speciesBits;
				return HashCommon.mix(yBits);
			}

			case field_11034:
			case field_11039: {
				final int xBits = (((modelState.posY() >> shift) & 0xFF) << 8) | ((modelState.posZ() >> shift) & 0xFF) | speciesBits;
				return HashCommon.mix(xBits);
			}

			case field_11043:
			case field_11035: {
				final int zBits = (((modelState.posX() >> shift) & 0xFF) << 8) | ((modelState.posY() >> shift) & 0xFF) | speciesBits;
				return HashCommon.mix(zBits);
			}

			default:
				return 0;
		}
	}

	/**
	 * Gives randomized (if applicable) texture rotation for the given face. If
	 * texture rotation type is FIXED, gives the textures default rotation. If
	 * texture rotation type is CONSISTENT, is based on species only. If texture
	 * rotation type is RANDOM, is based on position (chunked by texture size) and
	 * species (if applies).
	 */
	protected static TextureOrientation textureRotationForFace(class_2350 face, TextureSet tex, BaseModelState modelState) {
		switch (tex.transform()) {
			case ROTATE_180:
				return TextureOrientation.ROTATE_180;

			case ROTATE_270:
				return TextureOrientation.ROTATE_270;

			case ROTATE_90:
				return TextureOrientation.ROTATE_90;

			case ROTATE_BIGTEX: {
				final int species = modelState.hasSpecies() ? modelState.species() : 0;
				final ClockwiseRotation rot = species == 0 ? ClockwiseRotation.ROTATE_NONE : Useful.offsetEnumValue(ClockwiseRotation.ROTATE_NONE, HashCommon.mix(species) & 3);
				return TextureOrientation.find(rot, false, false);
			}

			case ROTATE_RANDOM:
				return TextureOrientation.find(Useful.offsetEnumValue(ClockwiseRotation.ROTATE_NONE, (textureHashForFace(face, tex, modelState) >> 8) & 3), false, false);

			case STONE_LIKE: {
				final int hash = textureHashForFace(face, tex, modelState);
				final ClockwiseRotation rot = (hash & 0b100000000) == 0 ? ClockwiseRotation.ROTATE_NONE : ClockwiseRotation.ROTATE_180;
				return TextureOrientation.find(rot, (hash & 0b1000000000) == 0, false);
			}

			case IDENTITY:
			default:
				return TextureOrientation.IDENTITY;
		}
	}
}
